@@ -117,6 +117,7 @@ copy_file (const char *realfilename1, const char *realfilename2)
     if (sourcefile == NULL)
 	{
 		fprintf (stderr, "can't open %s !\n", realfilename1);
+		fclose (targetfile);
 		return (-2);
 	}
 	while ((c = getc (sourcefile)) != EOF)
@@ -128,26 +129,38 @@ copy_file (const char *realfilename1, const char *realfilename2)
 }
 
 char*
-load_file(const char* realfilename)
+load_binary_file(const char* realfilename, long *file_size_return)
 {
 	struct stat st;
 	FILE* fp;
-	char* str;
-	int len;
+	char* data = NULL ;
 
 	/* Get the file size. */
 	if (stat(realfilename, &st)) return NULL;
-
 	/* Open the file. */
 	fp = fopen(realfilename, "rb");
-	if ( fp == NULL ) return NULL;
+	if ( fp != NULL ) 
+	{
+		long len ; 
+		/* Read in the file. */
+		data = safemalloc(st.st_size + 1);
+		len = fread(data, 1, st.st_size, fp);
+		if( file_size_return ) 
+			*file_size_return = len ; 
+		fclose(fp);
+	}
+	return data;
+}
 
-	/* Read in the file. */
-	str = NEW_ARRAY(char, st.st_size + 1);
-	len = fread(str, 1, st.st_size, fp);
-	if (len >= 0) str[len] = '\0';
+char*
+load_file(const char* realfilename)
+{
+	long len;
+	char* str = load_binary_file( realfilename, &len );
+
+	if (str != NULL && len >= 0) 
+		str[len] = '\0';
 
-	fclose(fp);
 	return str;
 }
 
@@ -177,6 +190,13 @@ make_file_name (const char *path, const char *file)
 	char         *filename;
 	int 		  len;
 
+	if( file == NULL ) 
+	{
+		if( path == NULL ) 
+			return NULL;	
+		return mystrdup(path);
+	}else if( path == NULL ) 
+		return mystrdup(file);			 
 	/* getting length */
 	for (ptr = (char *)path; ptr[i]; i++);
 	len = i+1;
@@ -208,11 +228,12 @@ put_file_home (const char *path_with_home)
 	if (path_with_home == NULL)
 		return NULL;
 	/* home dir ? */
-	if (path_with_home[0] != '~' || path_with_home[1] != '/')
-	{
-		char *t = mystrdup (path_with_home);
-		return t;
-	}
+	if ( strncmp(  path_with_home, "$HOME/", 6 ) == 0 )
+		path_with_home += 5 ;
+	else if (path_with_home[0] == '~' && path_with_home[1] == '/')
+		path_with_home += 1 ;
+	else
+		return mystrdup(path_with_home);
 
 	if (home == NULL)
 	{
@@ -221,9 +242,9 @@ put_file_home (const char *path_with_home)
 		home_len = strlen (home);
 	}
 
-	for (i = 2; path_with_home[i]; i++);
-	str = safemalloc (home_len + i);
-	for (ptr = str + home_len-1; i > 0; i--)
+	for (i = 0; path_with_home[i]; i++);
+	str = safemalloc (home_len + i + 1);
+	for (ptr = str + home_len; i >= 0; i--)
 		ptr[i] = path_with_home[i];
 	for (i = 0; i < home_len; i++)
 		str[i] = home[i];
@@ -252,11 +273,18 @@ find_file (const char *file, const char *pathlist, int type)
 	int            max_path = 0;
 	register char *ptr;
 	register int   i;
+	Bool local = False ;
 LOCAL_DEBUG_CALLER_OUT( "file \"%s\", pathlist = \"%s\"", file, pathlist );
 	if (file == NULL)
 		return NULL;
 
 	if (*file == '/' || *file == '~' || ((pathlist == NULL) || (*pathlist == '\0')))
+		local = True ;
+	else if( file[0] == '.' && (file[1] == '/' || (file[1] == '.' && file[2] == '/'))) 
+		local = True ;
+	else if( strncmp( file, "$HOME", 5) == 0 ) 
+		local = True ;
+	if( local ) 
 	{
 		path = put_file_home (file);
 		if ( access (path, type) == 0 )
@@ -278,16 +306,18 @@ LOCAL_DEBUG_CALLER_OUT( "file \"%s\", pathlist = \"%s\"", file, pathlist );
 			max_path = i;
 	}
 
-	path = safemalloc (max_path + 1 + len + 1 + 100);
+	path = safecalloc (1, max_path + 1 + len + 1);
 	strcpy( path+max_path+1, file );
 	path[max_path] = '/' ;
 
 	ptr = (char*)&(pathlist[0]) ;
 	while( ptr[0] != '\0' )
 	{
+		int skip ;
 		for( i = 0 ; ptr[i] == ':' ; ++i );
 		ptr += i ;
 		for( i = 0 ; ptr[i] != ':' && ptr[i] != '\0' ; ++i );
+		skip = i ;
 		if( i > 0 && ptr[i-1] == '/' )
 			i-- ;
 		if( i > 0 )
@@ -307,51 +337,59 @@ LOCAL_DEBUG_OUT( " found at: \"%s\"", res );
 				show_system_error(" looking for file %s:", file );
 #endif
 		}
-		if( ptr[i] == '/' )
-			ptr += i+1 ;
-		else
-			ptr += i ;
+		ptr += skip ;
 	}
 	free (path);
 	return NULL;
 }
 
-char         *
+static char *default_kdegnome_dir = "/usr" ;
+static char *default_kdehome_dir = "~/.kde" ;
+
+static char         *
 find_envvar (char *var_start, int *end_pos)
 {
-	char          backup, *name_start = var_start;
 	register int  i;
-	char         *var = NULL;
+	static char tmp[256];
+	char *result ;
 
 	if (var_start[0] == '{')
 	{
-		name_start++;
-		for (i = 1; var_start[i] && var_start[i] != '}'; i++);
+		for (i = 1; var_start[i] && var_start[i] != '}' && i < 255; i++)
+			tmp[i-1] = var_start[i] ;
+		tmp[i-1] = '\0' ;
 	} else
-		for (i = 0; isalnum ((int)var_start[i]) || var_start[i] == '_'; i++);
-
-	backup = var_start[i];
-	var_start[i] = '\0';
-	var = getenv (name_start);
-	var_start[i] = backup;
-
+	{	
+		for (i = 0; (isalnum ((int)var_start[i]) || var_start[i] == '_') && i < 255; i++)
+			tmp[i] = var_start[i] ;
+		tmp[i] = '\0';
+	}		
 	*end_pos = i;
-	if (backup == '}')
+	if (var_start[i] == '}')
 		(*end_pos)++;
-	return var;
+	
+	result = getenv (tmp);
+	if( result == NULL ) 
+	{
+		if( strcmp( tmp, "KDEDIR" ) == 0 || strcmp( tmp, "GNOMEDIR" ) == 0 ) 
+			result = default_kdegnome_dir ;
+		else if( strcmp( tmp, "KDEHOME" ) == 0 ) 
+			result = default_kdehome_dir ;
+	}
+	return result;
 }
 
 static char *
-do_replace_envvar (char *path)
+do_replace_envvar (const char *path)
 {
-	char         *data = path, *tmp;
+	char         *data = (char*)path, *tmp;
 	char         *home = getenv ("HOME");
 	int           pos = 0, len, home_len = 0;
 
 	if (path == NULL)
 		return NULL;
 	if (*path == '\0')
-		return path;
+		return (char*)path;
 	len = strlen (path);
 	if (home)
 		home_len = strlen (home);
@@ -421,7 +459,7 @@ replace_envvar (char **path)
 }
 
 char*
-copy_replace_envvar (char *path)
+copy_replace_envvar (const char *path)
 {
 	char         *res = do_replace_envvar( path );
 	return ( res == path )?mystrdup( res ):res;
@@ -432,9 +470,10 @@ copy_replace_envvar (char *path)
  * options to be passed
  */
 int
-is_executable_in_path (const char *name)
+get_executable_in_path (const char *name, char **fullname_return)
 {
 	static char  *cache = NULL;
+	static char  *cache_path = NULL;
 	static int    cache_result = 0, cache_len = 0, cache_size = 0;
 	static char  *env_path = NULL;
 	static int    max_path = 0;
@@ -447,6 +486,11 @@ is_executable_in_path (const char *name)
 			free (cache);
 			cache = NULL;
 		}
+		if (cache_path)
+		{
+			free (cache_path);
+			cache_path = NULL;
+		}
 		cache_size = 0;
 		cache_len = 0;
 		if (env_path)
@@ -472,7 +516,11 @@ is_executable_in_path (const char *name)
 
 	if (cache)
 		if (i == cache_len && strncmp (cache, name, i) == 0)
+		{
+			if (cache_result && fullname_return)
+				*fullname_return = mystrdup(cache_path);
 			return cache_result;
+		}
 
 	if (i > cache_size)
 	{
@@ -486,10 +534,17 @@ is_executable_in_path (const char *name)
 	strncpy (cache, name, i);
 	cache[i] = '\0';
 	cache_len = i;
-
+	if (cache_path)
+	{
+		free(cache_path);
+		cache_path = NULL;
+	}
+		
 	if (*cache == '/')
+	{
 		cache_result = (CheckFile (cache) == 0) ? 1 : 0;
-	else
+		cache_path = mystrdup (cache);
+	}else
 	{
 		char         *ptr, *path;
 		struct stat   st;
@@ -509,7 +564,7 @@ is_executable_in_path (const char *name)
 		}
 		path = safemalloc (max_path + cache_len + 2);
 		cache_result = 0;
-		for (ptr = env_path; *ptr && cache_result == 0; ptr += i)
+		for (ptr = env_path; *ptr ; ptr += i)
 		{
 			if (*ptr == ':')
 				ptr++;
@@ -519,14 +574,84 @@ is_executable_in_path (const char *name)
 			path[i + 1] = '\0';
 			strcat (path, cache);
 			if ((stat (path, &st) != -1) && (st.st_mode & S_IXUSR))
+			{
 				cache_result = 1;
+				cache_path = path;
+				path = NULL;
+				break;
+			}
 			LOCAL_DEBUG_OUT( "%s found \"%s\"", path, cache_result?"":"not" );
 		}
-		free (path);
+		if (path)
+			free (path);
 	}
+	
+	if (cache_result && fullname_return)
+		*fullname_return = mystrdup(cache_path);
 	return cache_result;
 }
 
+int
+is_executable_in_path (const char *name)
+{
+	return get_executable_in_path (name, NULL);
+}
+
+
+
+int
+my_scandir_ext ( const char *dirname, int (*filter_func) (const char *),
+				 Bool (*handle_direntry_func)( const char *fname, const char *fullname, struct stat *stat_info, void *aux_data), 
+				 void *aux_data)
+{
+	DIR          *d;
+	struct dirent *e;						   /* Pointer to static struct inside readdir() */
+	int           n = 0;					   /* Count of nl used so far */
+	char         *filename;					   /* For building filename to pass to stat */
+	char         *p;						   /* Place where filename starts */
+	struct stat   stat_info;
+
+	d = opendir (dirname);
+
+	if (d == NULL)
+		return -1;
+
+	filename = (char *)safemalloc (strlen (dirname) + PATH_MAX + 2);
+	if (filename == NULL)
+	{
+		closedir (d);
+		return -1;
+	}
+	strcpy (filename, dirname);
+	p = filename + strlen (filename);
+	if( *p != '/' )
+	{	
+		*p++ = '/';
+		*p = 0;									   /* Just in case... */
+	}
+	
+	while ((e = readdir (d)) != NULL)
+	{
+		if ((filter_func == NULL) || filter_func (&(e->d_name[0])))
+		{
+			/* Fill in the fields using stat() */
+			strcpy (p, e->d_name);
+			if (stat (filename, &stat_info) != -1)
+			{	
+				if( handle_direntry_func( e->d_name, filename, &stat_info, aux_data) )
+					n++;
+			}
+		}
+	}
+	free (filename);
+
+	if (closedir (d) == -1)
+		return -1;
+	/* Return the count of the entries */
+	return n;
+}
+
+
 /*
  * Non-NULL select and dcomp pointers are *NOT* tested, but should be OK.
  * They are not used by afterstep however, so this implementation should
@@ -534,9 +659,50 @@ is_executable_in_path (const char *name)
  *
  * c.ridd@isode.com
  */
+/* essentially duplicates whats above, 
+ * but for performance sake we don't want to merge them together :*/
+
+/* sort entries based on their type; directories come first */
+int
+direntry_compar_type (const struct direntry ** d1, const struct direntry ** d2)
+{
+	
+	return S_ISDIR ((**d2).d_mode) - S_ISDIR ((**d1).d_mode);
+}
+
+/* sort entries based on their names; A comes before Z */
+int
+direntry_compar_alpha (const struct direntry ** d1, const struct direntry ** d2)
+{
+	return strcmp ((**d1).d_name, (**d2).d_name);
+}
+
+int
+direntry_compar_type_alpha (const struct direntry ** d1, const struct direntry ** d2)
+{
+	int d = (S_ISDIR ((**d2).d_mode) - S_ISDIR ((**d1).d_mode)) ;
+	if( d == 0 ) 
+		return strcmp ((**d1).d_name, (**d2).d_name);
+	return d;
+}
+
+/* sort entries based on their mtimes; old entries before new entries */
+int
+direntry_compar_mtime (const struct direntry ** d1, const struct direntry ** d2)
+{
+	return (**d1).d_mtime - (**d2).d_mtime;
+}
+
+int
+direntry_compar_size (const struct direntry ** d1, const struct direntry ** d2)
+{
+	return (**d1).d_size - (**d2).d_size;
+}
+
+
 int
 my_scandir (char *dirname, struct direntry *(*namelist[]),
-			int (*select) (const char *), int (*dcomp) (struct direntry **, struct direntry **))
+			int (*filter_func) (const char *), my_sort_f dcomp)
 {
 	DIR          *d;
 	struct dirent *e;						   /* Pointer to static struct inside readdir() */
@@ -545,7 +711,6 @@ my_scandir (char *dirname, struct direntry *(*namelist[]),
 	int           n;						   /* Count of nl used so far */
 	int           sizenl;					   /* Number of entries in nl array */
 	int           j;
-	size_t        realsize;
 	char         *filename;					   /* For building filename to pass to stat */
 	char         *p;						   /* Place where filename starts */
 	struct stat   buf;
@@ -565,8 +730,11 @@ my_scandir (char *dirname, struct direntry *(*namelist[]),
 	}
 	strcpy (filename, dirname);
 	p = filename + strlen (filename);
-	*p++ = '/';
-	*p = 0;									   /* Just in case... */
+	if( *p != '/' )
+	{	
+		*p++ = '/';
+		*p = 0;									   /* Just in case... */
+	}
 
 	nl = NULL;
 	n = 0;
@@ -574,7 +742,7 @@ my_scandir (char *dirname, struct direntry *(*namelist[]),
 
 	while ((e = readdir (d)) != NULL)
 	{
-		if ((select == NULL) || select (&(e->d_name[0])))
+		if ((filter_func == NULL) || filter_func (&(e->d_name[0])))
 		{
 			/* add */
 			if (sizenl == n)
@@ -594,32 +762,18 @@ my_scandir (char *dirname, struct direntry *(*namelist[]),
 				}
 				nl = nnl;
 			}
-			realsize = offsetof (struct direntry, d_name)+strlen (e->d_name) + 1;
-			nl[n] = (struct direntry *)safemalloc (realsize);
-			if (nl[n] == NULL)
-			{
-				for (j = 0; j < n; j++)
-					free (nl[j]);
-				free (nl);
-				free (filename);
-				closedir (d);
-				return -1;
-			}
 			/* Fill in the fields using stat() */
 			strcpy (p, e->d_name);
-			if (stat (filename, &buf) == -1)
-			{
-				for (j = 0; j <= n; j++)
-					free (nl[j]);
-				free (nl);
-				free (filename);
-				closedir (d);
-				return -1;
+			if (stat (filename, &buf) != -1)
+			{	
+				size_t realsize = offsetof (struct direntry, d_name)+strlen (e->d_name) + 1;
+				nl[n] = (struct direntry *)safemalloc (realsize);
+				nl[n]->d_mode = buf.st_mode;
+				nl[n]->d_mtime = buf.st_mtime;
+				nl[n]->d_size  = buf.st_size;
+				strcpy (nl[n]->d_name, e->d_name);
+				n++;
 			}
-			nl[n]->d_mode = buf.st_mode;
-			nl[n]->d_mtime = buf.st_mtime;
-			strcpy (nl[n]->d_name, e->d_name);
-			n++;
 		}
 	}
 	free (filename);
@@ -654,7 +808,7 @@ my_scandir (char *dirname, struct direntry *(*namelist[]),
 }
 
 /*
- * Use this function as the select argument to my_scandir to make it ignore
+ * Use this function as the filter_func argument to my_scandir to make it ignore
  * all files and directories starting with "."
  */
 int
@@ -666,8 +820,25 @@ ignore_dots (const char *d_name)
 int
 no_dots_except_include (const char *d_name)
 {
-	return !(d_name[0] == '.' && mystrcasecmp (d_name, ".include"));
+	if( d_name[0] != '.' || mystrcasecmp (d_name, ".include") == 0 )
+	{	
+		register  int i = 0;
+		while( d_name[i] != '\0' ) ++i ; 
+		return (i > 0 && d_name[i-1] != '~');
+	}
+	return False;
+}
+
+int
+no_dots_except_directory (const char *d_name)
+{
+	if( d_name[0] != '.' || mystrcasecmp (d_name, ".directory") == 0 )
+	{	
+		register  int i = 0;
+		while( d_name[i] != '\0' ) ++i ; 
+		return (i > 0 && d_name[i-1] != '~');
+	}
+	return False;
 }
 
 
-
